import pandas as pd

import stripe
from typing import Text, List, Dict, Any
from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import INSERTQueryParser, DELETEQueryParser, UPDATEQueryParser, DELETEQueryExecutor, UPDATEQueryExecutor

from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.utilities.handlers.query_utilities.select_query_utilities import SELECTQueryParser, SELECTQueryExecutor


from mindsdb.utilities import log

logger = log.getLogger(__name__)


class CustomersTable(APITable):
    """The Stripe Customers Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Customer data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Customers matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'customers',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        customers_df = pd.json_normalize(self.get_customers(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            customers_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        customers_df = select_statement_executor.execute_query()

        return customers_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_customers(limit=1)).columns.tolist()

    def get_customers(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        customers = stripe.Customer.list(**kwargs)
        return [customer.to_dict() for customer in customers]


class ProductsTable(APITable):
    """The Stripe Products Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Product data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Products matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'products',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            products_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        products_df = select_statement_executor.execute_query()

        return products_df

    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into Stripe "POST v1/products" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['id', 'name', 'active', 'description', 'metadata'],
            mandatory_columns=['name'],
            all_mandatory=False,
        )
        product_data = insert_statement_parser.parse_query()
        self.create_products(product_data)

    def update(self, query: ast.Update) -> None:
        """
        Updates data from Stripe "POST v1/products/:id" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products())
        update_query_executor = UPDATEQueryExecutor(
            products_df,
            where_conditions
        )

        products_df = update_query_executor.execute_query()
        product_ids = products_df['id'].tolist()
        self.update_products(product_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from Stripe "DELETE v1/products/:id" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products())
        delete_query_executor = DELETEQueryExecutor(
            products_df,
            where_conditions
        )

        products_df = delete_query_executor.execute_query()
        product_ids = products_df['id'].tolist()
        self.delete_products(product_ids)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_products(limit=1)).columns.tolist()

    def get_products(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        products = stripe.Product.list(**kwargs)
        return [product.to_dict() for product in products]

    def create_products(self, product_data: List[Dict[Text, Any]]) -> None:
        stripe = self.handler.connect()
        for product in product_data:
            created_product = stripe.Product.create(**product)
            if 'id' not in created_product.to_dict():
                raise Exception('Product creation failed')
            else:
                logger.info(f'Product {created_product.to_dict()["id"]} created')

    def update_products(self, product_ids: List[Text], values_to_update: Dict[Text, Any]) -> None:
        stripe = self.handler.connect()
        for product_id in product_ids:
            updated_product = stripe.Product.modify(product_id, **values_to_update)
            if 'id' not in updated_product.to_dict():
                raise Exception('Product update failed')
            else:
                logger.info(f'Product {updated_product.to_dict()["id"]} updated')

    def delete_products(self, product_ids: List[Text]) -> None:
        stripe = self.handler.connect()
        for product_id in product_ids:
            deleted_product = stripe.Product.delete(product_id)
            if 'id' not in deleted_product.to_dict():
                raise Exception('Product deletion failed')
            else:
                logger.info(f'Product {deleted_product.to_dict()["id"]} deleted')


class PaymentIntentsTable(APITable):
    """The Stripe Payment Intents Table implementation"""
    
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Payment Intents data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Payment Intents matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'payment_intents',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        payment_intents_df = pd.json_normalize(self.get_payment_intents(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            payment_intents_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        payment_intents_df = select_statement_executor.execute_query()

        return payment_intents_df
    def delete(self, query: ast.Delete) -> None:
        """
        Cancels Stripe Payment Intents and updates the local data.

        Parameters
        ----------
        query : ast.Delete
        Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
        If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        if 'payment_intents_df' not in self.__dict__:
            self.payment_intents_df = pd.json_normalize(self.get_payment_intents())

        delete_query_executor = DELETEQueryExecutor(
            self.payment_intents_df,
            where_conditions
        )

        canceled_payment_intents_df = delete_query_executor.execute_query()

        payment_intent_ids = canceled_payment_intents_df['id'].tolist()
        self.cancel_payment_intents(payment_intent_ids)

        
        self.payment_intents_df = self.payment_intents_df[~self.payment_intents_df['id'].isin(payment_intent_ids)]

    def cancel_payment_intents(self, payment_intent_ids: List[str]) -> None:
        stripe = self.handler.connect()
        for payment_intent_id in payment_intent_ids:
            try:
              
                payment_intent = stripe.PaymentIntent.retrieve(payment_intent_id)
                if payment_intent.status in ['requires_payment_method', 'requires_capture', 'requires_confirmation', 'requires_action', 'processing']:
                    stripe.PaymentIntent.cancel(payment_intent_id)
                else:
                     logger.warning(f"Payment intent {payment_intent_id} is in status {payment_intent.status} and cannot be canceled.")
            except stripe.error.StripeError as e:
                logger.error(f"Error cancelling payment intent {payment_intent_id}: {str(e)}")
                
    def update(self, query: 'ast.Update') -> None:
        """
        Updates data in Stripe "POST /v1/payment_intents/:id" API endpoint.

        Parameters
        ----------
        query : ast.Update
            Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        payment_intents_df = pd.json_normalize(self.get_payment_intents())
        update_query_executor = UPDATEQueryExecutor(
            payment_intents_df,
            where_conditions
        )

        payment_intents_df = update_query_executor.execute_query()
        payment_intent_ids = payment_intents_df['id'].tolist()
        self.update_payment_intents(payment_intent_ids, values_to_update)

    def update_payment_intents(self, payment_intent_ids: list, values_to_update: dict) -> None:
        for payment_intent_id in payment_intent_ids:
            stripe.PaymentIntent.modify(payment_intent_id, **values_to_update)


    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_payment_intents(limit=1)).columns.tolist()

    def get_payment_intents(self, **kwargs) -> List[Dict]:
        payment_intents = stripe.PaymentIntent.list(**kwargs)
        return [payment_intent for payment_intent in payment_intents.data]

    
    
    def insert(self, query: 'ast.Insert') -> None:
        """
        Inserts data into Stripe "POST /v1/payment_intents" API endpoint.

        Parameters
        ----------
        query : ast.Insert
            Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['amount', 'currency', 'description', 'payment_method_types'],
            mandatory_columns=['amount', 'currency'],
            all_mandatory=True
        )
        payment_intent_data = insert_statement_parser.parse_query()
        self.create_payment_intent(payment_intent_data)

    def create_payment_intent(self, payment_intent_data: list) -> None:
        for data in payment_intent_data:
            stripe.PaymentIntent.create(**data)
        
    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_payment_intents(limit=1)).columns.tolist()

    def get_payment_intents(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        payment_intents = stripe.PaymentIntent.list(**kwargs)
        return [payment_intent.to_dict() for payment_intent in payment_intents]

      
class RefundsTable(APITable):
    """The Stripe Refund Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Refund data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Refunds matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'refunds',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        refunds_df = pd.json_normalize(self.get_refunds(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            refunds_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        refunds_df = select_statement_executor.execute_query()

        return refunds_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_refunds(limit=1)).columns.tolist()

    def get_refunds(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        refunds = stripe.Refund.list(**kwargs)
        return [refund.to_dict() for refund in refunds ]
     
    
class PayoutsTable(APITable):
    """The Stripe Payouts Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls Stripe Payout data.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Stripe Payouts matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'payouts',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        payouts_df = pd.json_normalize(self.get_payouts(limit=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            payouts_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        payouts_df = select_statement_executor.execute_query()

        return payouts_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_payouts(limit=1)).columns.tolist()

    def get_payouts(self, **kwargs) -> List[Dict]:
        stripe = self.handler.connect()
        payouts = stripe.Payout.list(**kwargs)
        return [payout.to_dict() for payout in payouts]



import json
import os
import shutil
from pathlib import Path

from mindsdb.integrations.handlers.file_handler import Handler as FileHandler
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


class FileController:
    def __init__(self):
        self.config = Config()
        self.fs_store = FsStore()
        self.dir = os.path.join(self.config.paths["content"], "file")

    def get_files_names(self):
        """return list of files names"""
        return [
            x[0]
            for x in db.session.query(db.File.name).filter_by(company_id=ctx.company_id)
        ]

    def get_file_meta(self, name):
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            return None
        columns = file_record.columns
        if isinstance(columns, str):
            columns = json.loads(columns)
        return {
            "name": file_record.name,
            "columns": columns,
            "row_count": file_record.row_count,
        }

    def get_files(self):
        """Get list of files

        Returns:
            list[dict]: files metadata
        """
        file_records = (
            db.session.query(db.File).filter_by(company_id=ctx.company_id).all()
        )
        files_metadata = [
            {
                "name": record.name,
                "row_count": record.row_count,
                "columns": record.columns,
            }
            for record in file_records
        ]
        return files_metadata

    def save_file(self, name, file_path, file_name=None):
        """Save the file to our store

        Args:
            name (str): with that name file will be available in sql api
            file_name (str): file name
            file_path (str): path to the file

        Returns:
            int: id of 'file' record in db
        """
        files_metadata = self.get_files()
        if name in [x["name"] for x in files_metadata]:
            raise Exception(f"File already exists: {name}")

        if file_name is None:
            file_name = Path(file_path).name

        file_dir = None
        try:
            df, _col_map = FileHandler._handle_source(file_path)

            ds_meta = {"row_count": len(df), "column_names": list(df.columns)}

            file_record = db.File(
                name=name,
                company_id=ctx.company_id,
                source_file_path=file_name,
                file_path="",
                row_count=ds_meta["row_count"],
                columns=ds_meta["column_names"],
            )
            db.session.add(file_record)
            db.session.commit()
            store_file_path = f"file_{ctx.company_id}_{file_record.id}"
            file_record.file_path = store_file_path
            db.session.commit()

            file_dir = Path(self.dir).joinpath(store_file_path)
            file_dir.mkdir(parents=True, exist_ok=True)
            source = file_dir.joinpath(file_name)
            # NOTE may be delay between db record exists and file is really in folder
            shutil.move(file_path, str(source))

            self.fs_store.put(store_file_path, base_dir=self.dir)
        except Exception as e:
            logger.error(e)
            if file_dir is not None:
                shutil.rmtree(file_dir)
            raise

        return file_record.id

    def delete_file(self, name):
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            return None
        file_id = file_record.id
        db.session.delete(file_record)
        db.session.commit()
        self.fs_store.delete(f"file_{ctx.company_id}_{file_id}")
        return True

    def get_file_path(self, name):
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            raise Exception(f"File '{name}' does not exists")
        file_dir = f"file_{ctx.company_id}_{file_record.id}"
        self.fs_store.get(file_dir, base_dir=self.dir)
        return str(
            Path(self.dir)
            .joinpath(file_dir)
            .joinpath(Path(file_record.source_file_path).name)
        )



from typing import List

import pandas as pd
from mindsdb_sql_parser import ast

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities import log

from mindsdb_sql_parser import ast

logger = log.getLogger(__name__)


class ConfluencePagesTable(APITable):
    """Confluence Pages Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Confluence "get_all_pages_from_space" API endpoint"""
        conditions = extract_comparison_conditions(query.where)
        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 50

        space_name = None
        page_id = None

        for a_where in conditions:
            if a_where[1] == "space":
                space_name = a_where[2]
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for space")
            elif a_where[1] == "id":
                page_id = a_where[2]
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for id")
            elif a_where[1] not in ["id", "space"]:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        if space_name is None:
            raise ValueError("Space name must be provided in the WHERE clause")

        if page_id is not None:
            try:
                page = self.handler.connect().get_page_by_id(page_id, expand='body.storage,space')
                confluence_pages_records = [page] if page['space']['key'] == space_name else []
            except Exception as e:
                logger.error(f"Error fetching page with ID {page_id}: {str(e)}")
                confluence_pages_records = []
        else:
            confluence_pages_records = self.handler.connect().get_all_pages_from_space(
                space_name, start=0, limit=total_results, expand="body.storage,space"
            )

        if not confluence_pages_records:
            return pd.DataFrame(columns=self.get_columns())

        confluence_pages_df = pd.json_normalize(confluence_pages_records)
        
        confluence_pages_df.columns = confluence_pages_df.columns.str.replace("body.storage.value", "body")
        confluence_pages_df['space'] = confluence_pages_df['space.key']
        
        confluence_pages_df = confluence_pages_df[self.get_columns()]

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                col = target.parts[-1]
                if col in self.get_columns():
                    selected_columns.append(col)
                else:
                    raise ValueError(f"Unknown column: {col}")
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        confluence_pages_df = confluence_pages_df[selected_columns]

        if query.order_by and len(query.order_by) > 0:
            sort_columns = []
            sort_ascending = []
            for an_order in query.order_by:
                if isinstance(an_order.field, ast.Identifier):
                    column = an_order.field.parts[-1]
                    if column in selected_columns:
                        sort_columns.append(column)
                        sort_ascending.append(an_order.direction == "ASC")
                    else:
                        raise ValueError(f"Order by unknown column {column}")
                else:
                    raise ValueError(f"Unsupported order by clause: {an_order}")
            
            if sort_columns:
                confluence_pages_df = confluence_pages_df.sort_values(
                    by=sort_columns,
                    ascending=sort_ascending
                )

        return confluence_pages_df
    
    
    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "id",
            "type",
            "status",
            "title",
            "space",
            "body",
            "_links.self",
            "_links.webui",
        ]

    def insert(self, query: ast.Insert):
        """Inserts a new page into the Confluence space

        Parameters
        ----------
        query : ast.Insert
            Given SQL INSERT query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        columns = [col.name for col in query.columns]

        for i, val in enumerate(query.values):
            params = dict(zip(columns, val))

            self.handler.connect().create_page(
                space=params["space"],
                title=params["title"],
                body=params["body"],
            )

    def update(self, query: ast.Update):
        """Updates a page in the Confluence space

        Parameters
        ----------
        query : ast.Update
            Given SQL UPDATE query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        params = extract_comparison_conditions(query.where)
        title = query.update_columns["title"].get_string()
        body = query.update_columns["body"].get_string()

        for param in params:
            if param[1] == "id":
                id = param[2]
                self.handler.connect().update_page(
                    page_id=id,
                    title=title,
                    body=body,
                )

    def delete(self, query: ast.Delete):
        """Deletes a page from the Confluence space

        Parameters
        ----------
        query : ast.Delete
            Given SQL DELETE query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        params = extract_comparison_conditions(query.where)
        for param in params:
            if param[1] == "id":
                id = param[2]
                self.handler.connect().remove_page(
                    page_id=id,
                )



import io
import pandas as pd
from box_sdk_gen import BoxClient, BoxDeveloperTokenAuth, CreateFolderParent, UploadFileAttributes, UploadFileAttributesParentField
from box_sdk_gen.internal import utils
from typing import Dict, Optional, Text

from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Select, Identifier, Insert

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.integrations.libs.api_handler import APIHandler, APIResource


class ListFilesTable(APIResource):

    def list(self, conditions=None, limit=None, sort=None, targets=None, **kwargs):
        files = self.handler._list_files()
        data = []
        for file in files:
            item = {
                "path": file["path"],
                "name": file["name"],
                "extension": file["extension"],
            }
            data.append(item)
        df = pd.DataFrame(data)
        return df

    def get_columns(self):
        return ["path", "name", "extension"]


class FileTable(APIResource):

    def _get_file_df(self):
        try:
            df = self.handler._read_file(self.table_name)
            if df is None:
                raise Exception(f"No such file found for the path: {self.box_path}")

            return df
        except Exception as e:
            self.handler.logger.error(e)

    def list(self, conditions=None, limit=None, sort=None, targets=None, **kwargs):
        return self._get_file_df()

    def get_columns(self):
        df = self.handler._read_file(self.table_name)
        return df.columns.tolist()

    def insert(self, query: Insert) -> None:
        columns = [col.name for col in query.columns]
        data = [dict(zip(columns, row)) for row in query.values]
        df_new = pd.DataFrame(data)
        df_existing = self._get_file_df()
        df_combined = pd.concat([df_existing, df_new], ignore_index=True)
        self.handler._write_file(self.table_name, df_combined)


class BoxHandler(APIHandler):

    name = "box"
    supported_file_formats = ["csv", "tsv", "json", "parquet"]

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs):
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.logger = log.getLogger(__name__)
        self.client = None
        self.is_connected = False
        self._files_table = ListFilesTable(self)
        self._register_table("files", self._files_table)

    def connect(self):
        try:
            if self.is_connected:
                return
            if "token" not in self.connection_data:
                raise ValueError("Developer token must be provided.")
            auth = BoxDeveloperTokenAuth(token=self.connection_data["token"])
            self.client = BoxClient(auth=auth)
            self.client.folders.get_folder_items("0", limit=1)
            self.is_connected = True
            self.logger.info("Connected to Box")
        except Exception as e:
            self.logger.error(f"Error connecting to Box: {e}")

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)
        try:
            self.connect()
            response.success = True
        except Exception as e:
            self.logger.error(f"Error with Box Handler while establish connection: {e}")
            response.error_message = str(e)
        return response

    def disconnect(self):
        if not self.is_connected:
            return
        self.client = None
        self.is_connected = False
        self.logger.info("Disconnected from Box")

    def _read_as_content(self, file_path) -> utils.Buffer:
        """
        Read files as content
        """
        try:
            id = "0"  # Root folder id. The value is always "0".

            items = file_path.strip('/').split('/')

            for item in items:
                item_id = None
                for entry in self.client.folders.get_folder_items(id).entries:
                    if entry.name == item:
                        item_id = entry.id
                        break
                if item_id:
                    id = item_id
                else:
                    raise ValueError(f"{item} not found. Specify the correct path.")

            downloaded_file_content = self.client.downloads.download_file(id)
            buffer = utils.read_byte_stream(downloaded_file_content)
            return buffer
        except Exception as e:
            self.logger.error(f"Error when downloading a file from Box: {e}")

    def query(self, query: ASTNode) -> Response:

        if isinstance(query, Select):
            table_name = query.from_table.parts[-1]
            if table_name == "files":
                table = self._files_table
                df = table.select(query)
                has_content = False
                for target in query.targets:
                    if (
                        isinstance(target, Identifier)
                        and target.parts[-1].lower() == "content"
                    ):
                        has_content = True
                        break
                if has_content:
                    df["content"] = df["path"].apply(self._read_as_content)
            else:
                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            return Response(RESPONSE_TYPE.TABLE, data_frame=df)
        elif isinstance(query, Insert):
            table_name = query.table.parts[-1]
            table = FileTable(self, table_name=table_name)
            table.insert(query)
            return Response(RESPONSE_TYPE.OK)
        else:
            raise NotImplementedError(
                "Only SELECT and INSERT operations are supported."
            )

    def get_tables(self) -> Response:
        table_names = list(self._tables.keys())
        df = pd.DataFrame(table_names, columns=["table_name"])
        return Response(RESPONSE_TYPE.TABLE, data_frame=df)

    def get_columns(self, table_name: str) -> Response:
        table = self._get_table(Identifier(table_name))
        columns = table.get_columns()
        df = pd.DataFrame(columns, columns=["column_name"])
        return Response(RESPONSE_TYPE.TABLE, data_frame=df)

    def list_files_in_folder(self, folder_id, files=[]):
        items = self.client.folders.get_folder_items(folder_id).entries
        for item in items:
            if item.type == 'folder':
                self.list_files_in_folder(item.id, files)
            elif item.type == 'file':
                extension = item.name.split(".")[-1].lower()
                if extension in self.supported_file_formats:
                    file = self.client.files.get_file_by_id(item.id)
                    full_path = [path.name for path in file.path_collection.entries][1:]
                    full_path.append(item.name)
                    files.append(
                        {
                            "path": "/".join(full_path),
                            "name": item.name,
                            "extension": extension,
                        }
                    )

        return files

    def _list_files(self, path=""):
        return self.list_files_in_folder("0")

    def _read_file(self, path) -> pd.DataFrame:
        try:
            buffer = self._read_as_content(path)
            extension = path.split(".")[-1].lower()
            if extension == "csv":
                df = pd.read_csv(io.BytesIO(buffer))
            elif extension == "tsv":
                df = pd.read_csv(io.BytesIO(buffer), sep="\t")
            elif extension == "json":
                df = pd.read_json(io.BytesIO(buffer))
            elif extension == "parquet":
                df = pd.read_parquet(io.BytesIO(buffer))
            else:
                raise ValueError(f"Unsupported file format: {extension}")
            return df
        except Exception as e:
            self.logger.error(f"Error with Box Handler while reading file: {e}")

    def upload_file(self, path, buffer):
        parent_folder_id = "0"

        directories = path.strip('/').split('/')

        for direc in directories[:-1]:
            sub_folder_id = None
            for item in self.client.folders.get_folder_items(parent_folder_id).entries:
                if item.type == "folder" and item.name == direc:
                    sub_folder_id = item.id
                    break
            if sub_folder_id:
                parent_folder_id = sub_folder_id
            else:
                new_sub_folder = self.client.folders.create_folder(direc, CreateFolderParent(id=parent_folder_id))
                parent_folder_id = new_sub_folder.id
                self.logger.debug(f"folder {direc} not available in path. Creating {direc}...")

        self.client.uploads.upload_file(UploadFileAttributes(name=directories[-1], parent=UploadFileAttributesParentField(id=parent_folder_id)), buffer)

    def _write_file(self, path, df: pd.DataFrame):
        extension = path.split(".")[-1].lower()
        buffer = io.BytesIO()
        if extension == "csv":
            df.to_csv(buffer, index=False)
        elif extension == "tsv":
            df.to_csv(buffer, index=False, sep="\t")
        elif extension == "json":
            df.to_json(buffer, orient="records")
        elif extension == "parquet":
            df.to_parquet(buffer, index=False)
        else:
            raise ValueError(f"Unsupported file format: {extension}")
        buffer.seek(0)
        self.upload_file(path, buffer)



import unittest
from unittest.mock import patch, MagicMock
from mindsdb.integrations.handlers.ckan_handler.ckan_handler import CkanHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb_sql_parser import ast


class CkanHandlerTest(unittest.TestCase):
    def setUp(self):
        self.handler = CkanHandler('test_ckan_handler', connection_data={
            "url": "http://mock-ckan-url.com",
            "api_key": "mock_api_key"
        })
        self.patcher = patch('mindsdb.integrations.handlers.ckan_handler.ckan_handler.RemoteCKAN')
        self.mock_ckan = self.patcher.start()
        self.mock_ckan_instance = MagicMock()
        self.mock_ckan.return_value = self.mock_ckan_instance

    def tearDown(self):
        self.patcher.stop()

    def test_check_connection(self):
        self.mock_ckan_instance.action.site_read.return_value = True
        response = self.handler.check_connection()
        self.assertTrue(response.success)

    def test_get_tables(self):
        self.mock_ckan_instance.action.package_list.return_value = [
            "package1",
            "package2",
        ]
        result = self.handler.get_tables()
        self.assertIsNotNone(result)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)

    def test_query_package_ids(self):
        mock_packages = [
            {
                "id": "pkg1",
                "name": "Package 1",
                "title": "Test Package 1",
                "num_resources": 2,
            },
            {
                "id": "pkg2",
                "name": "Package 2",
                "title": "Test Package 2",
                "num_resources": 1,
            },
        ]
        self.mock_ckan_instance.action.package_search.return_value = {
            "results": mock_packages
        }

        query = ast.Select(
            targets=[ast.Star()],
            from_table=ast.Identifier("package_ids"),
            limit=ast.Constant(10),
        )
        result = self.handler.query(query)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 2)

    def test_query_resource_ids(self):
        mock_packages = [
            {
                "id": "pkg1",
                "resources": [
                    {
                        "id": "res1",
                        "name": "Resource 1",
                        "format": "CSV",
                        "datastore_active": True,
                    },
                    {
                        "id": "res2",
                        "name": "Resource 2",
                        "format": "XSLX",
                        "datastore_active": False,
                    },
                ],
            },
            {
                "id": "pkg2",
                "resources": [
                    {
                        "id": "res3",
                        "name": "Resource 3",
                        "format": "CSV",
                        "datastore_active": True,
                    }
                ],
            },
        ]
        self.mock_ckan_instance.action.package_search.return_value = {
            "results": mock_packages
        }

        query = ast.Select(
            targets=[ast.Star()],
            from_table=ast.Identifier("resource_ids"),
            limit=ast.Constant(10),
        )
        result = self.handler.query(query)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 2)

    def test_query_datastore_without_resource_id(self):
        query = ast.Select(
            targets=[ast.Star()],
            from_table=ast.Identifier("datastore"),
            limit=ast.Constant(10),
        )
        result = self.handler.query(query)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 1)
        self.assertIn("message", result.data_frame.columns)

    def test_query_datastore_with_resource_id(self):
        mock_records = [
            {"id": 1, "name": "Record 1"},
            {"id": 2, "name": "Record 2"},
            {"id": 3, "name": "Record 3"},
        ]
        self.mock_ckan_instance.action.datastore_search.return_value = {
            "records": mock_records
        }

        query = ast.Select(
            targets=[ast.Star()],
            from_table=ast.Identifier("datastore"),
            where=ast.BinaryOperation(
                "=",
                args=[ast.Identifier("resource_id"), ast.Constant("test_resource_id")],
            ),
            limit=ast.Constant(10),
        )
        result = self.handler.query(query)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 3)

    def test_native_query(self):
        self.mock_ckan_instance.action.package_list.return_value = [
            "package1",
            "package2",
            "package3",
        ]

        query = "package_list:"
        result = self.handler.native_query(query)
        self.assertEqual(result.type, RESPONSE_TYPE.TABLE)
        self.assertEqual(len(result.data_frame), 3)


if __name__ == "__main__":
    unittest.main()



from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log
import duckdb
import pandas as pd

from azure.storage.blob import BlobServiceClient

from contextlib import contextmanager
from typing import List, Text, Optional, Dict
from mindsdb.integrations.libs.api_handler import APIResource, APIHandler
from mindsdb.integrations.utilities.sql_utils import FilterCondition
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser.ast import Select, Identifier, Insert
from mindsdb_sql_parser import parse_sql

logger = log.getLogger(__name__)


class ListFilesTable(APIResource):

    def list(self,
             targets: List[str] = None,
             conditions: List[FilterCondition] = None,
             *args, **kwargs) -> pd.DataFrame:

        tables = self.handler.get_files()
        data = []
        for path in tables:
            path = path.replace('`', '')
            item = {
                'path': path,
                'name': path[path.rfind('/') + 1:],
                'extension': path[path.rfind('.') + 1:]
            }

            data.append(item)

        return pd.DataFrame(data=data, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        return ["path", "name", "extension", "content"]


class FileTable(APIResource):

    def list(self, targets: List[str] = None, table_name=None, *args, **kwargs) -> pd.DataFrame:
        return self.handler.read_as_table(table_name)

    def add(self, data, table_name=None):
        df = pd.DataFrame(data)
        return self.handler.add_data_to_table(table_name, df)


class AzureBlobHandler(APIHandler):
    """
    This handler handles connection and execution of the SQL statements on Azure Blob.
    """

    name = "azureblob"
    supported_file_formats = ['csv', 'tsv', 'json', 'parquet']

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs):
        super().__init__(name)
        """ constructor
        Args:
            name (str): the handler name
        """

        self.connection = None
        self.is_connected = False
        self._tables = {}
        self._files_table = ListFilesTable(self)
        self.container_name = None

        self.connection_data = connection_data

        if 'container_name' in connection_data:
            self.container_name = connection_data['container_name']

        if 'connection_string' in connection_data:
            self.connection_string = connection_data['connection_string']

    def connect(self) -> BlobServiceClient:
        """ Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.connection

        blob_service_client = BlobServiceClient.from_connection_string(conn_str=self.connection_string)

        self.connection = blob_service_client
        self.is_connected = True
        return blob_service_client

    def check_connection(self) -> StatusResponse:
        """ Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            client = self.connect()
            client.get_account_information()
            response.success = True

        except Exception as e:
            logger.error(f'Error connecting to Azure Blob: {e}!')
            response.error_message = e

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def disconnect(self):
        """
        Closes the connection to the Azure Blob account if it's currently open.
        """
        if not self.is_connected:
            return
        self.connection.close()
        self.is_connected = False

    @contextmanager
    def _connect_duckdb(self):
        """
        Creates temporal duckdb database which is able to connect to the Azure Blob account.
        Have to be used as context manager

        Returns:
            DuckDBPyConnection
        """
        # Connect to Azure Blob via DuckDB.
        duckdb_conn = duckdb.connect(":memory:")
        duckdb_conn.execute('INSTALL azure')
        duckdb_conn.execute('LOAD azure')

        # Configure mandatory credentials.
        duckdb_conn.execute(f'SET azure_storage_connection_string="{self.connection_string}"')

        try:
            yield duckdb_conn
        finally:
            duckdb_conn.close()

    def read_as_table(self, key) -> pd.DataFrame:
        """
        Read object as dataframe. Uses duckdb
        """

        with self._connect_duckdb() as connection:
            cursor = connection.execute(f'SELECT * FROM "azure://{self.container_name}/{key}"')
            return cursor.fetchdf()

    def _read_as_content(self, key) -> None:
        """
        Read object as content
        """

        connection = self.connect()
        client = connection.get_blob_client(container=self.container_name, blob=key)

        return client.download_blob()

    def add_data_to_table(self, key, df) -> None:
        pass
        """
        Writes the table to a file in the azure container.

        Raises:
            CatalogException: If the table does not exist in the DuckDB connection.
        """

        # Check if the file exists in the Container.

        try:
            client = self.connect()
            blob_client = client.get_blob_client(container=self.container_name, blob=key)
            blob_client.close()

        except Exception as e:
            logger.error(f'Error querying the file {key} in the container {self.container_name}, {e}!')
            raise e

        with self._connect_duckdb() as connection:
            # copy
            connection.execute(f'CREATE TABLE tmp_table AS SELECT * FROM "azure://{self.container_name}/{key}"')

            # insert
            connection.execute("INSERT INTO tmp_table BY NAME SELECT * FROM df")

            # upload
            connection.execute(f"COPY tmp_table TO 'azure://{self.container_name}/{key}'")

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        query_ast = parse_sql(query)
        return self.query(query_ast)

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Raises:
            ValueError: If the file format is not supported or the file does not exist in the Azure Blob Container.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """

        self.connect()

        if isinstance(query, Select):
            table_name = query.from_table.parts[-1].replace('`', '')

            if table_name == 'files':
                table = self._files_table
                df = table.select(query)

                # add content
                has_content = False
                for target in query.targets:
                    if isinstance(target, Identifier) and target.parts[-1].lower() == 'content':
                        has_content = True
                        break
                if has_content:
                    df['content'] = df['path'].apply(self._read_as_content)
            else:
                extension = table_name.split('.')[-1]
                if extension not in self.supported_file_formats:
                    logger.error(f'The file format {extension} is not supported!')
                    raise ValueError(f'The file format {extension} is not supported!')

                table = FileTable(self, table_name=table_name)
                df = table.select(query)

            response = Response(
                RESPONSE_TYPE.TABLE,
                data_frame=df
            )
        elif isinstance(query, Insert):
            table_name = query.table.parts[-1]
            table = FileTable(self, table_name=table_name)
            table.insert(query)
            response = Response(RESPONSE_TYPE.OK)
        else:
            raise NotImplementedError

        return response

    def get_files(self) -> List[str]:
        client = self.connect()
        container_client = client.get_container_client(self.container_name)
        all_files = container_client.list_blobs()

        # Wrap the object names with backticks to prevent SQL syntax errors.
        supported_files = [
            f"`{file.get('name')}`"
            for file in all_files if file.get('name').split('.')[-1] in self.supported_file_formats
        ]

        return supported_files

    def get_tables(self) -> Response:
        """
        Retrieves a list of tables (objects) in the Azure Containers.

        Each object is considered a table. Only the supported file formats are considered as tables.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """

        # Get only the supported file formats.
        # Wrap the object names with backticks to prevent SQL syntax errors.

        supported_files = self.get_files()

        # virtual table with list of files
        supported_files.insert(0, 'files')

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                supported_files,
                columns=['table_name']
            )
        )

        return response

    def get_columns(self, table_name: str) -> Response:
        """
        Retrieves column details for a specified table (object) in the Azure Blob Container.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        """
        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        query = f"SELECT * FROM {table_name} LIMIT 5"

        result = self.query(query)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    'column_name': result.data_frame.columns,
                    'data_type': [data_type if data_type != 'object' else 'string' for data_type in result.data_frame.dtypes]
                }
            )
        )

        return response



from __future__ import annotations

import asyncio
import logging
import math
import os
from typing import Any, Dict, List, Optional, Sequence, Tuple
from uuid import uuid4
from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain_core.callbacks import Callbacks

from mindsdb.integrations.utilities.rag.settings import DEFAULT_RERANKING_MODEL, DEFAULT_LLM_ENDPOINT
from langchain_core.documents import Document
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_openai import ChatOpenAI


log = logging.getLogger(__name__)


class LLMReranker(BaseDocumentCompressor):
    filtering_threshold: float = 0.5  # Default threshold for filtering
    model: str = DEFAULT_RERANKING_MODEL  # Model to use for reranking
    temperature: float = 0.0  # Temperature for the model
    openai_api_key: Optional[str] = None
    remove_irrelevant: bool = True  # New flag to control removal of irrelevant documents,
    base_url: str = DEFAULT_LLM_ENDPOINT
    num_docs_to_keep: Optional[int] = None  # How many of the top documents to keep after reranking & compressing.

    _api_key_var: str = "OPENAI_API_KEY"
    client: Optional[Any] = None

    class Config:
        arbitrary_types_allowed = True

    async def search_relevancy(self, query: str, document: str) -> Any:
        openai_api_key = self.openai_api_key or os.getenv(self._api_key_var)

        # Initialize the ChatOpenAI client
        client = ChatOpenAI(openai_api_base=self.base_url, api_key=openai_api_key, model=self.model, temperature=0,
                            logprobs=True)

        # Create the message history for the conversation
        message_history = [
            SystemMessage(
                content="""Your task is to classify whether the document is relevant to the search query provided below. Answer just "YES" or "NO"."""),
            HumanMessage(content=f"""Document: ```{document}```; Search query: ```{query}```""")
        ]

        # Generate the response using LangChain's chat model
        response = await client.agenerate(
            messages=[message_history],
            max_tokens=1
        )

        # Return the response from the model
        return response.generations[0]

    async def _rank(self, query_document_pairs: List[Tuple[str, str]]) -> List[Tuple[str, float]]:
        # Gather results asynchronously for all query-document pairs
        results = await asyncio.gather(
            *[self.search_relevancy(query=query, document=document) for (query, document) in query_document_pairs]
        )

        ranked_results = []

        for idx, result in enumerate(results):
            # Extract the log probability (assuming logprobs are provided in LangChain response)
            msg = result[0].message
            logprob = msg.response_metadata['logprobs']['content'][0]['logprob']
            prob = math.exp(logprob)
            answer = result[0].message.content  # The model's "YES" or "NO" response

            # Calculate the score based on the model's response
            if answer == "YES":
                score = prob
            elif answer.lower().strip().startswith("y"):
                score = prob
            elif answer == "NO":
                score = 1 - prob
            elif answer.lower().strip().startswith("n"):
                score = 1 - prob
            else:
                score = 0.0  # Default if something unexpected happens

            # Append the document and score to the result
            ranked_results.append((query_document_pairs[idx][1], score))  # (document, score)

        return ranked_results

    def compress_documents(
            self,
            documents: Sequence[Document],
            query: str,
            callbacks: Optional[Callbacks] = None,
    ) -> Sequence[Document]:
        """Compress documents using OpenAI's rerank capability with individual document assessment."""
        log.info(f"Compressing documents. Initial count: {len(documents)}")
        if len(documents) == 0:
            log.warning("No documents to compress. Returning empty list.")
            return []

        doc_contents = [doc.page_content for doc in documents]
        query_documents_pairs = [(query, doc) for doc in doc_contents]

        # Create event loop and run async code
        import asyncio
        try:
            rankings = asyncio.get_event_loop().run_until_complete(self._rank(query_documents_pairs))
        except RuntimeError:
            # If no event loop is available, create a new one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            rankings = loop.run_until_complete(self._rank(query_documents_pairs))

        compressed = []
        for ind, ranking in enumerate(rankings):
            doc = documents[ind]
            document_text, score = ranking
            doc.metadata["relevance_score"] = score
            doc.metadata["is_relevant"] = score > self.filtering_threshold
            # Add the document to the compressed list if it is relevant or if we are not removing irrelevant documents
            if not self.remove_irrelevant:
                compressed.append(doc)
            elif doc.metadata["is_relevant"]:
                compressed.append(doc)

        log.info(f"Compression complete. {len(compressed)} documents returned")
        if not compressed:
            log.warning("No documents found after compression")

        if self.num_docs_to_keep is not None:
            # Sort by relevance score with highest first.
            compressed.sort(
                key=lambda d: d.metadata.get('relevance_score', 0) if d.metadata else 0,
                reverse=True
            )
            compressed = compressed[:self.num_docs_to_keep]

        # Handle retrieval callbacks to account for reranked & compressed docs.
        callbacks = callbacks if callbacks else []
        run_id = uuid4().hex
        if not isinstance(callbacks, list):
            callbacks = callbacks.handlers
        for callback in callbacks:
            callback.on_retriever_end(compressed, run_id=run_id)
        return compressed

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {
            "model": self.model,
            "temperature": self.temperature,
            "remove_irrelevant": self.remove_irrelevant,
        }



import os
from typing import Dict

from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities.config import Config

"""Contains utilities to be used by handlers."""


def get_api_key(
    api_name: str,
    create_args: Dict[str, str],
    engine_storage: HandlerStorage = None,
    strict: bool = True,
):
    """Gets the API key needed to use an ML Handler.

    Args:
        api_name (str): Name of the API (e.g. openai, anthropic)
        create_args (Dict[str, str]): Args user passed to the created model with USING keyword
        engine_storage (HandlerStorage): Engine storage for the ML handler
        strict (bool): Whether or not to require the API key

    Returns:
        api_key (str): The API key

    API_KEY preference order:
        1. provided at inference
        2. provided at model creation
        3. provided at engine creation
        4. api key env variable
        5. api_key setting in config.json
    """
    # Special case for vLLM - always return dummy key
    if api_name == "vllm":
        return "EMPTY"

    # 1
    if "using" in create_args and f"{api_name.lower()}_api_key" in create_args["using"]:
        return create_args["using"][f"{api_name.lower()}_api_key"]

    # 2
    if f"{api_name.lower()}_api_key" in create_args:
        return create_args[f"{api_name.lower()}_api_key"]

    # 3
    if engine_storage is not None:
        connection_args = engine_storage.get_connection_args()
        if f"{api_name.lower()}_api_key" in connection_args:
            return connection_args[f"{api_name.lower()}_api_key"]

    # 4
    api_key = os.getenv(f"{api_name.lower()}_api_key")
    if api_key is not None:
        return api_key
    api_key = os.getenv(f"{api_name.upper()}_API_KEY")
    if api_key is not None:
        return api_key

    # 5
    config = Config()
    api_cfg = config.get(api_name, {})
    if f"{api_name.lower()}_api_key" in api_cfg:
        return api_cfg[f"{api_name.lower()}_api_key"]

    if strict:
        raise Exception(
            f"Missing API key '{api_name.lower()}_api_key'. Either re-create this ML_ENGINE specifying the '{api_name.lower()}_api_key' parameter, or re-create this model and pass the API key with `USING` syntax."
        )  # noqa
    return None



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    aws_access_key_id={
        'type': ARG_TYPE.STR,
        'description': 'The access key for the AWS account.'
    },
    aws_secret_access_key={
        'type': ARG_TYPE.STR,
        'description': 'The secret key for the AWS account.',
        'secret': True
    },
    region_name={
        'type': ARG_TYPE.STR,
        'description': 'The AWS region where the Athena tables are created.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The name of the Athena database.'
    },
    workgroup={
        'type': ARG_TYPE.STR,
        'description': 'The Athena Workgroup'
    },
    catalog={
        'type': ARG_TYPE.STR,
        'description': 'The AWS Data Catalog'
    },
    results_output_location={
        'type': ARG_TYPE.STR,
        'description': 'The Athena Query Results Output Location s3://bucket-path/athena-query-results'
    },
    check_interval={
        'type': ARG_TYPE.INT,
        'description': 'The interval in seconds to check Athena for query results. Default is 0 seconds.'
    }
)

connection_args_example = OrderedDict(
    aws_access_key_id='<default>',
    aws_secret_access_key='<default>',
    region_name='us-east-1',
    catalog='AwsDataCatalog',
    database='default',
    workgroup='primary',
    results_output_location='s3://<bucket>/athena-query-results/',
    check_interval=0
)
